import logging
import pandas as pd
import os
from dotenv import load_dotenv
from sqlalchemy.types import Integer, Float, String, Boolean, DateTime, TypeEngine, BigInteger

load_dotenv(override=True)

def disable_logging(func):
    def wrapper(*args, **kwargs):
        logging.disable(logging.CRITICAL)
        try:
            return func(*args, **kwargs)
        finally:
            logging.disable(logging.NOTSET)
    return wrapper

def map_types(dtype: pd.api.types) -> TypeEngine:
    if pd.api.types.is_integer_dtype(dtype):
        # Check if dtype is 64-bit integer
        if dtype == pd.Int64Dtype() or dtype == 'int64':
            return BigInteger()
        else:
            return Integer()
    elif pd.api.types.is_float_dtype(dtype):
        return Float()
    elif pd.api.types.is_bool_dtype(dtype):
        return Boolean()
    elif pd.api.types.is_datetime64_any_dtype(dtype):
        return DateTime()
    else:
        return String()
    
def create_default_uri_from_env():
    if all([os.getenv("POSTGRES_DB_USERNAME"), os.getenv("POSTGRES_DB_PASSWORD"), os.getenv("POSTGRES_DB_HOST"), os.getenv("POSTGRES_DB_PORT"), os.getenv("POSTGRES_DB_DEFAULT_NAME")]):
        return (
            f"postgresql://{os.getenv('POSTGRES_DB_USERNAME')}:{os.getenv('POSTGRES_DB_PASSWORD')}@{os.getenv('POSTGRES_DB_HOST')}:"
            f"{os.getenv('POSTGRES_DB_PORT')}/{os.getenv('POSTGRES_DB_DEFAULT_NAME')}?sslmode={os.getenv('POSTGRES_DB_SSLMODE', 'require')}"
        )
    else:
        return None


import logging
from decimal import Decimal
from typing import Dict, Optional, Any
import pandas as pd

class QueryHandling:
    def _query_cleaning(self, query: str) -> str:
        """
        Clean and normalize the input query by replacing specific words.

        Args:
            query (str): The input query string.

        Returns:
            str: The cleaned and normalized query string.
        """
        break_words = {
            'maximum': 'max',
            'minimum': 'min',
        }
        query = query.lower()
        for word, replacement in break_words.items():
            query = query.replace(word, replacement)
        return query


    def _make_json_serializable(self, resp):
        if 'data' in resp and isinstance(resp['data'], list):
            for record in resp['data']:
                for key, value in record.items():
                    if isinstance(value, pd.Timestamp):
                        record[key] = value.isoformat()
        return resp

    def query(self, query: str, generative: bool = False, n_retries: int = 2) -> Dict[str, Optional[Any]]:
        """
        Process the input query and generate an SQL query, execute it on the database, 
        and optionally generate a generative response.

        Args:
            query (str): The input query string.
            generative (bool): Flag to indicate if a generative response is required.
            n_retries (int): The number of retry attempts for generating and executing the SQL query.

        Returns:
            dict: A dictionary containing the results of the query execution.
        """
        result = {'success': False, 'message': None, 'sql': None, 'data': None, 'generation': None}
        query = self._query_cleaning(query)
        if n_retries < 0 or not isinstance(n_retries, int):
            logging.info("The number of retries must be a positive integer. Defaulting to 2 retries.")
            n_retries = 2
        
        retry_sql = None
        error_message = None
        for i in range(n_retries + 1):
            if i > 0:
                retry_prompt = (
                    f"The user query was: '{query}'\n\n"
                    f"The SQL query generated from OpenAI in the first attempt is: '{retry_sql}'.\n\n"
                    f"This SQL failed to execute due to the following error:\n\n'{error_message}'.\n\n"
                    f"Please try to generate the appropriate SQL again, taking into account the error message. Make adjustments based on the above error message.\n\n"
                )
                prompt = self._build_sql_prompt(query, starter_prompt=retry_prompt)
            else:
                prompt = self._build_sql_prompt(query)
                logging.info(f"Generating SQL for query: '{query}'...")
            
            sql_dict = self._sql_openai_call(prompt)
            if not sql_dict:
                logging.error("Failed to generate SQL query, trying again.")
                error_message = "Failed to generate SQL query."
                continue
            
            sql = sql_dict.get('sql')
            in_domain = sql_dict.get('in_domain')
            if not in_domain:
                logging.error("Question is not in related to the data, please ask a different question.")
                result['message'] = "Query is not in domain."
                return result
            db_result_dict = self._query_db(sql)
            if not db_result_dict.get('success'):
                # logging.error(f"Error message: {db_result_dict.get('message')}")
                # logging.error(f"SQL: '{sql}'")
                retry_sql = sql
                error_message = db_result_dict.get('message')
                logging.error("Couldn't execute SQL on the database. Tweaking the query and trying again...")
                continue
            
            result_df = db_result_dict.get('data')
            if result_df.empty:
                logging.info("Query returned no results.")
                error_message = "Query returned no results."
                continue
            
            logging.info("Query executed successfully.")
            for col in result_df.columns:
                if result_df[col].dtype == object and isinstance(result_df[col].iloc[0], Decimal):
                    result_df[col] = result_df[col].astype(float)
            
            result['success'] = True
            result['sql'] = sql
            result['message'] = "Query executed successfully."
            result['data'] = result_df.to_dict(orient='records')

            if generative:
                generative_prompt = self._build_generative_prompt(query, result_df)
                result['generation'] = self._generative_openai_call(generative_prompt)

            return self._make_json_serializable(result)

        result['message'] = "Failed to execute query after retries."
        return result


import os
import logging
from typing import Optional
from sqlalchemy import create_engine, exc, text
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from dotenv import load_dotenv
from .database_management import DatabaseManagement
from .llm_integration import LLMIntegration
from .query_handling import QueryHandling
from .utils import create_default_uri_from_env
from urllib.parse import urlparse, urlunparse

load_dotenv(override=True)
logging.basicConfig(level=logging.INFO, format='%(message)s')

class PostgresDB(DatabaseManagement, LLMIntegration, QueryHandling):
    """
    A class to manage PostgreSQL database connections and operations,
    including database creation, connection, and deletion.

    Inherits from DatabaseManagement, LLMIntegration, and QueryHandling.
    """

    def __init__(
        self, 
        name: str = 'db', 
        postgres_connection_string: Optional[str] = None,
    ) -> None:
        """
        Initialize the PostgresDB instance.

        Args:
            name (str): The name of the database.
            postgres_connection_string (Optional[str]): The PostgreSQL connection string.

        Raises:
            ValueError: If the PostgreSQL connection string is not provided.
            Exception: For any other exceptions during initialization.
        """
        try:
            self.name = name

            if not postgres_connection_string:
                postgres_connection_string = (
                    os.getenv('POSTGRES_CONNECTION_STRING') or create_default_uri_from_env()
                )

            if not postgres_connection_string:
                raise ValueError("Missing parameters to construct the PostgreSQL connection string.")
            
            self.postgres_connection_string: str = postgres_connection_string
            self.default_engine: Engine = create_engine(self.postgres_connection_string, isolation_level="AUTOCOMMIT")

            self._check_connection(engine=self.default_engine)

            if self._database_exists():
                logging.info(f"The database '{self.name}' already exists...")
            else:
                logging.info(f"The database '{self.name}' does not exist. Creating database...")
                self._create_database()
            
            self._create_db_engine()
            self._check_connection(engine=self.engine)

        except Exception as e:
            logging.error(f"Error initializing PostgresDB: {e}")
            raise

    def _check_connection(self, engine: Engine) -> None:
        """
        Check if the connection to the PostgreSQL database is successful.
        
        Raises:
            Exception: If the connection cannot be established.
        """
        try:
            if engine == self.default_engine:
                message = "Successfully connected to the default PostgreSQL database..."
            elif engine == self.engine:
                message = f"Successfully connected to database '{self.name}'..."
            with engine.connect() as connection:
                logging.info(message)
        except exc.SQLAlchemyError as e:
            logging.error(f"Error connecting to the PostgreSQL database: {e}")
            raise

    def _database_exists(self) -> bool:
        """
        Check if a database with the given name exists.

        Returns:
            bool: True if the database exists, False otherwise.
        """
        query = text("SELECT 1 FROM pg_database WHERE datname = :dbname")
        try:
            with self.default_engine.connect() as connection:
                result = connection.execute(query, {'dbname': self.name})
                exists = result.scalar() is not None
                return exists
        except SQLAlchemyError as e:
            logging.error(f"Error checking if database '{self.name}' exists: {e}")
            raise

    def _create_database(self) -> None:
        """
        Create a new database with the given name.

        Raises:
            Exception: If there is an error creating the database.
        """
        create_db_query = text(f"CREATE DATABASE {self.name}")
        try:
            with self.default_engine.connect() as connection:
                connection.execute(create_db_query)
                logging.info(f"Database '{self.name}' created successfully...")
        except SQLAlchemyError as e:
            logging.error(f"Error creating database '{self.name}': {e}")
            raise
    
    def _create_db_engine(self) -> str:
        """
        Create a new SQLAlchemy engine URI for the specified database.

        Args:
            dbname (str): The name of the database.

        Returns:
            str: The SQLAlchemy engine URI for the specified database.
        """
        try:
            parsed_url = urlparse(self.postgres_connection_string)
            new_path = f"/{self.name}"
            new_url = parsed_url._replace(path=new_path)
            self.db_uri = urlunparse(new_url)            
            self.engine = create_engine(self.db_uri, isolation_level="AUTOCOMMIT")
        except SQLAlchemyError as e:
            logging.error(f"Error creating engine for database '{self.name}': {e}")
            raise

    def _disconnect_all_users(self, dbname: str) -> None:
        """
        Disconnect all users from the specified database.

        Args:
            dbname (str): The name of the database from which to disconnect users.

        Raises:
            Exception: If there is an error disconnecting the users.
        """
        disconnect_query = text(f"""
            SELECT pg_terminate_backend(pg_stat_activity.pid)
            FROM pg_stat_activity
            WHERE pg_stat_activity.datname = :dbname
              AND pid <> pg_backend_pid();
        """)
        try:
            with self.default_engine.connect() as connection:
                connection.execute(disconnect_query, {'dbname': dbname})
        except SQLAlchemyError as e:
            logging.error(f"Error disconnecting users from database '{dbname}': {e}")
            raise

    def delete_database(self, confirm: bool = True) -> None:
        """
        Delete a database with the given name.

        Args:
            confirm (bool): Whether to ask for confirmation before deleting the database.

        Raises:
            Exception: If there is an error deleting the database.
        """
        delete_db_query = text(f"DROP DATABASE {self.name}")
        try:
            if confirm:
                confirm_input = input(f"Are you sure you want to delete the database '{self.name}'? (y/n): ")
                if confirm_input.lower() != 'y':
                    logging.info("Database deletion aborted.")
                    return

            self._disconnect_all_users(self.name)
            with self.default_engine.connect() as connection:
                connection.execute(delete_db_query)
                logging.info(f"Database '{self.name}' deleted successfully...")
        except SQLAlchemyError as e:
            logging.error(f"Error deleting database '{self.name}': {e}")
            raise



import hashlib
import logging
import warnings
import os
import re
import pandas as pd
from sqlalchemy import create_engine, exc, text, Column, MetaData, Table as SQLTable
from sqlalchemy.dialects.postgresql import insert
import chardet
from .utils import disable_logging, map_types
import time

class DatabaseManagement:
    def __init__(self, engine):
        self.engine = engine

    def create_tables(self, data_directory: str) -> None:
        """
        Create or update tables from CSV or Excel files in the specified directory.

        Args:
            data_directory (str): The directory containing the data files.
        """
        for file_name in os.listdir(data_directory):
            if file_name.endswith('.csv') or file_name.endswith('.xlsx'):
                table_name = os.path.splitext(file_name)[0].lower().strip().replace(' ', '_')
                file_path = os.path.join(data_directory, file_name)
                logging.info(f"Creating or updating table '{table_name}' from file '{file_name}'...")
                try:
                    self._create_or_update_table(table_name, file_path)
                except Exception as e:
                    # logging.error(f"Error creating or updating table '{table_name}'...")
                    logging.error(f"Error creating or updating table '{table_name}'")
                    with open('error.log', 'a') as log_file:
                        log_file.write(f"Error creating or updating table '{table_name}':\n\n{str(e)}\n")




    def _create_or_update_table(self, table_name: str, file_path: str) -> None:
        @disable_logging
        def _clean_column_names(df: pd.DataFrame) -> pd.DataFrame:
            df.columns = [
                re.sub(r'[^a-z0-9_]', '', col.replace('%', '_percent').replace('+', '_plus').replace('-', '_minus').replace('/', '_per_').strip().lower().replace(' ', '_'))
                for col in df.columns
            ]
            return df

        @disable_logging
        def _convert_datetime_columns(df: pd.DataFrame) -> pd.DataFrame:
            def _is_datetime_column(column):
                if column.dtype != object:
                    return False
                date_patterns = [
                    r'\d{2}/\d{2}/\d{2}',
                    r'\d{2}/\d{2}/\d{4}',
                    r'\d{4}/\d{2}/\d{2}',
                    r'\d{2}-\d{2}-\d{4}',
                    r'\d{4}-\d{2}-\d{2}',
                    r'\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2}',
                    r'\d{4}/\d{2}/\d{2} \d{2}:\d{2}:\d{2}',
                    r'\d{2}-\d{2}-\d{4} \d{2}:\d{2}:\d{2}',
                    r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}',
                ]
                for pattern in date_patterns:
                    if column.str.match(pattern).any():
                        try:
                            with warnings.catch_warnings():
                                warnings.simplefilter("ignore", UserWarning)
                                pd.to_datetime(column, errors='raise', dayfirst=True)
                            return True
                        except (ValueError, TypeError):
                            continue
                return False

            datetime_columns = [col for col in df.columns if _is_datetime_column(df[col])]
            
            for column in datetime_columns:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    df[column] = pd.to_datetime(df[column], errors='coerce', infer_datetime_format=True)
            
            return df

        try:
            start_time = time.time()

            with open(file_path, 'rb') as file:
                raw_data = file.read()
                result = chardet.detect(raw_data)
                encoding = result['encoding']

            if file_path.endswith('.csv'):
                data = pd.read_csv(file_path, encoding=encoding, low_memory=False)
            elif file_path.endswith('.xlsx'):
                data = pd.read_excel(file_path, engine='openpyxl')
            else:
                logging.error(f"Unsupported file format for file '{file_path}'.")
                return

            data = _clean_column_names(data)
            data = _convert_datetime_columns(data)

            metadata = MetaData()
            columns = [Column(column_name, map_types(dtype)) for column_name, dtype in data.dtypes.items()]
            table = SQLTable(table_name, metadata, *columns)
            metadata.create_all(self.engine)

            with self.engine.connect() as connection:
                try:
                    existing_data = pd.read_sql_table(table_name, connection)
                except Exception:
                    existing_data = pd.DataFrame()

            if not existing_data.empty:
                existing_data['_hash'] = existing_data.apply(lambda row: hashlib.md5(pd.util.hash_pandas_object(row, index=True).values).hexdigest(), axis=1)
                data['_hash'] = data.apply(lambda row: hashlib.md5(pd.util.hash_pandas_object(row, index=True).values).hexdigest(), axis=1)

                non_duplicate_data = data[~data['_hash'].isin(existing_data['_hash'])].drop(columns=['_hash'])
            else:
                non_duplicate_data = data

            if not non_duplicate_data.empty:
                logging.info(f"Inserting {len(non_duplicate_data)} new rows into '{table_name}'...")
                non_duplicate_data.to_sql(table_name, self.engine, if_exists='append', index=False, method='multi', chunksize=1000)
                logging.info(f"New data has been successfully inserted for table '{table_name}'.")

            end_time = time.time()
            elapsed_time = end_time - start_time
            logging.info(f"Time taken to process and insert data for table '{table_name}': {elapsed_time:.2f} seconds")

        except Exception as e:
            logging.error(f"Error processing file '{file_path}'...")
            raise

    def delete_tables(self, tables: list = [], all: bool = False, confirm: bool = True) -> None:
        """
        Delete the specified tables or all tables if 'all' is set to True.

        Args:
            tables (list): List of table names to delete.
            all (bool): Flag to indicate if all tables should be deleted.
            confirm (bool): Flag to indicate if confirmation is required before deleting.
        """
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        tables = tables if not all else list(metadata.tables.keys())

        def _delete_table(connection, table):
            try:
                table = metadata.tables[table]
                logging.info(f"Dropping table '{table}'...")
                table.drop(connection)
                logging.info(f"Table {table} dropped successfully.")
            except Exception as e:
                logging.error(f"Error deleting table {table}...")

        with self.engine.begin() as connection:
            for table in tables:
                if confirm:
                    response = input(f"Are you sure you want to delete table '{table}'? This action can't be undone. (y/n): ")
                    if response.lower().strip() == 'y':
                        _delete_table(connection, table)
                    else:
                        logging.info(f"Table '{table}' deletion aborted.")
                else:
                    _delete_table(connection, table)

    def check_primary_key(self, table_name: str) -> None:
        """
        Check if the specified table has a primary key.

        Args:
            table_name (str): The name of the table to check.
        """
        with self.engine.connect() as con:
            try:
                result = con.execute(text(f"""
                    SELECT kcu.column_name
                    FROM information_schema.table_constraints tc
                    JOIN information_schema.key_column_usage kcu
                    ON tc.constraint_name = kcu.constraint_name
                    WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'PRIMARY KEY';
                """))
                primary_keys = result.fetchall()
                result_dict = {'exists': len(primary_keys) > 0, 'field': primary_keys[0][0] if primary_keys else None}
                logging.info(f"Primary key check for table '{table_name}': {result_dict}")
            except Exception as e:
                logging.error(f"Error checking primary key for table '{table_name}'...")
                raise

    def add_primary_key(self, table_name: str, field: str) -> None:
        """
        Add a primary key to the specified table.

        Args:
            table_name (str): The name of the table.
            field (str): The name of the field to set as the primary key.
        """
        with self.engine.connect() as con:
            try:
                result = con.execute(text(f"""
                    SELECT kcu.column_name
                    FROM information_schema.table_constraints tc
                    JOIN information_schema.key_column_usage kcu
                    ON tc.constraint_name = kcu.constraint_name
                    WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'PRIMARY KEY';
                """))
                existing_primary_keys = [row[0] for row in result.fetchall()]

                if len(existing_primary_keys) == 1 and existing_primary_keys[0] == field:
                    return

                fk_constraints = []

                for existing_pk in existing_primary_keys:
                    fk_result = con.execute(text(f"""
                        SELECT tc.table_name, kcu.column_name, ccu.table_name AS foreign_table_name, ccu.column_name AS foreign_column_name, tc.constraint_name
                        FROM information_schema.table_constraints AS tc 
                        JOIN information_schema.key_column_usage AS kcu
                        ON tc.constraint_name = kcu.constraint_name
                        JOIN information_schema.constraint_column_usage AS ccu
                        ON ccu.constraint_name = tc.constraint_name
                        WHERE tc.constraint_type = 'FOREIGN KEY' AND ccu.table_name='{table_name}' AND ccu.column_name='{existing_pk}';
                    """))
                    fk_constraints = fk_result.fetchall()

                    for fk in fk_constraints:
                        fk_constraint_name = fk[4]
                        fk_table_name = fk[0]
                        con.execute(text(f"""
                            ALTER TABLE {fk_table_name} DROP CONSTRAINT {fk_constraint_name};
                        """))
                        logging.info(f"Dropped foreign key constraint '{fk_constraint_name}' from table '{fk_table_name}'.")

                    con.execute(text(f"""
                        ALTER TABLE {table_name} DROP CONSTRAINT {existing_pk};
                    """))
                    logging.info(f"Dropped existing primary key constraint '{existing_pk}' from table '{table_name}'.")

                con.execute(text(f"""
                    ALTER TABLE {table_name}
                    ADD PRIMARY KEY ({field});
                """))
                logging.info(f"Primary key '{field}' added to table '{table_name}'.")

                for fk in fk_constraints:
                    fk_constraint_name = fk[4]
                    fk_table_name = fk[0]
                    fk_column_name = fk[1]
                    foreign_table_name = fk[2]
                    foreign_column_name = fk[3]
                    con.execute(text(f"""
                        ALTER TABLE {fk_table_name}
                        ADD CONSTRAINT {fk_constraint_name}
                        FOREIGN KEY ({fk_column_name})
                        REFERENCES {foreign_table_name} ({foreign_column_name});
                    """))
                    logging.info(f"Recreated foreign key constraint '{fk_constraint_name}' on table '{fk_table_name}'.")
                con.commit()
            except Exception as e:
                logging.error(f"Failed to update primary key '{field}' for table '{table_name}'...")
                raise

    def drop_primary_key(self, table_name: str) -> None:
        """
        Drop the primary key from the specified table.

        Args:
            table_name (str): The name of the table.
        """
        with self.engine.connect() as con:
            try:
                result = con.execute(text(f"""
                    SELECT kcu.column_name
                    FROM information_schema.table_constraints tc
                    JOIN information_schema.key_column_usage kcu
                    ON tc.constraint_name = kcu.constraint_name
                    WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'PRIMARY KEY';
                """))
                existing_primary_keys = [row[0] for row in result.fetchall()]

                for existing_pk in existing_primary_keys:
                    con.execute(text(f"""
                        ALTER TABLE {table_name} DROP CONSTRAINT {existing_pk};
                    """))
                    logging.info(f"Dropped existing primary key constraint '{existing_pk}' from table '{table_name}'.")
            except Exception as e:
                logging.error(f"Failed to drop primary key constraints from table '{table_name}'...")
                raise

    def check_foreign_keys(self, table_name: str) -> bool:
        """
        Check if the specified table has foreign keys.

        Args:
            table_name (str): The name of the table to check.

        Returns:
            bool: True if the table has foreign keys, False otherwise.
        """
        with self.engine.connect() as con:
            try:
                result = con.execute(text(f"""
                    SELECT tc.constraint_name
                    FROM information_schema.table_constraints tc
                    JOIN information_schema.key_column_usage kcu
                    ON tc.constraint_name = kcu.constraint_name
                    WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'FOREIGN KEY';
                """))
                foreign_keys = result.fetchall()
                result_dict = {'exists': len(foreign_keys) > 0, 'constraints': [fk[0] for fk in foreign_keys]}
                logging.info(f"Foreign key check for table '{table_name}': {result_dict}")
                return result_dict['exists']
            except Exception as e:
                logging.error(f"Error checking foreign keys for table '{table_name}'...")
                raise

    def add_foreign_key(self, table_name: str, field: str, reference_table: str, reference_field: str) -> None:
        """
        Add a foreign key to the specified table.

        Args:
            table_name (str): The name of the table.
            field (str): The field to set as the foreign key.
            reference_table (str): The referenced table.
            reference_field (str): The referenced field.
        """
        with self.engine.connect() as con:
            try:
                self.add_primary_key(reference_table, reference_field)

                result = con.execute(text(f"""
                    SELECT tc.constraint_name
                    FROM information_schema.table_constraints tc
                    JOIN information_schema.key_column_usage kcu
                    ON tc.constraint_name = kcu.constraint_name
                    WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'FOREIGN KEY' AND kcu.column_name = '{field}';
                """))
                existing_constraints = result.fetchall()

                for constraint in existing_constraints:
                    constraint_name = constraint[0]
                    con.execute(text(f"""
                        ALTER TABLE {table_name} DROP CONSTRAINT {constraint_name};
                    """))
                    logging.info(f"Dropped existing foreign key constraint '{constraint_name}' from table '{table_name}'.")

                con.execute(text(f"""
                    ALTER TABLE {table_name}
                    ADD CONSTRAINT fk_{table_name}_{field}
                    FOREIGN KEY ({field})
                    REFERENCES {reference_table} ({reference_field});
                """))
                con.commit()
                logging.info(f"Foreign key '{field}' added to table '{table_name}' referencing '{reference_table}({reference_field})'.")
            except Exception as e:
                logging.error(f"Failed to update foreign key '{field}' for table '{table_name}'...")
                raise

    def drop_foreign_key(self, table_name: str, field: str = None, all: bool = False) -> None:
        """
        Drop the foreign key from the specified table.

        Args:
            table_name (str): The name of the table.
            field (str): The field to drop the foreign key from.
            all (bool): Flag to indicate if all foreign keys should be dropped.
        """
        if not field and not all:
            logging.error("Please provide a field or set 'all' to True to drop all foreign key constraints.")
            return
        with self.engine.begin() as connection:
            try:
                if all:
                    result = connection.execute(text(f"""
                        SELECT tc.constraint_name
                        FROM information_schema.table_constraints tc
                        JOIN information_schema.key_column_usage kcu
                        ON tc.constraint_name = kcu.constraint_name
                        WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'FOREIGN KEY';
                    """))
                    foreign_keys = result.fetchall()
                    for fk in foreign_keys:
                        connection.execute(text(f"ALTER TABLE {table_name} DROP CONSTRAINT {fk[0]}"))
                        logging.info(f"Dropped foreign key constraint '{fk[0]}' from table '{table_name}'.")
                else:
                    result = connection.execute(text(f"""
                        SELECT tc.constraint_name
                        FROM information_schema.table_constraints tc
                        JOIN information_schema.key_column_usage kcu
                        ON tc.constraint_name = kcu.constraint_name
                        WHERE tc.table_name = '{table_name}' AND tc.constraint_type = 'FOREIGN KEY' AND kcu.column_name = '{field}';
                    """))
                    foreign_keys = result.fetchall()
                    for fk in foreign_keys:
                        connection.execute(text(f"ALTER TABLE {table_name} DROP CONSTRAINT {fk[0]}"))
                        logging.info(f"Dropped foreign key constraint '{fk[0]}' from table '{table_name}'.")
            except Exception as e:
                logging.error(f"Failed to drop foreign key constraints from table '{table_name}'...")
                raise

    def _query_db(self, sql: str) -> dict:
        """
        Execute the specified SQL query on the database.

        Args:
            sql (str): The SQL query to execute.

        Returns:
            dict: The result of the query execution.
        """
        with self.engine.connect() as con:
            try:
                result = con.execute(text(sql))
                data = result.fetchall()
                columns = result.keys()
                df = pd.DataFrame(data, columns=columns)
                if df.empty:
                    return {'success': False, 'data': df, 'message': "No data returned."}
                else:
                    return {'success': True, 'data': df, 'message': None}
            except Exception as e:
                # logging.error(f"Error querying the database...")
                return {'success': False, 'data': None, 'message': str(e)}



import logging
import re
import pandas as pd
from sqlalchemy import text
from .utils import disable_logging
from openai import OpenAI

client = OpenAI()

class LLMIntegration:
    def _build_sql_prompt(self, query: str, starter_prompt: str = '') -> str:
        """
        Build a SQL prompt for generating SQL queries from a given query string.

        Args:
            query (str): The input query string.
            starter_prompt (str): An optional starter prompt to include additional context.

        Returns:
            str: The constructed prompt string.
        """
        try:
            prompt = starter_prompt
            prompt += f"""You are a Postgres SQL expert who has been tasked with writing a SQL query to extract data from the database based on this query: '{query}'."""
            with self.engine.connect() as con:
                columns_result = con.execute(text("""
                    SELECT table_name, column_name, data_type
                    FROM information_schema.columns
                    WHERE table_schema = 'public'
                    ORDER BY table_name, column_name;
                """))
                table_columns = columns_result.fetchall()

                primary_keys_result = con.execute(text("""
                    SELECT kcu.table_name, tco.constraint_name, kcu.column_name 
                    FROM information_schema.table_constraints tco
                    JOIN information_schema.key_column_usage kcu 
                        ON kcu.constraint_name = tco.constraint_name
                        AND kcu.constraint_schema = tco.constraint_schema
                        AND kcu.constraint_name = tco.constraint_name
                    WHERE tco.constraint_type = 'PRIMARY KEY' AND kcu.table_schema = 'public'
                    ORDER BY kcu.table_name, kcu.column_name;
                """))
                primary_keys = primary_keys_result.fetchall()

                foreign_keys_result = con.execute(text("""
                    SELECT 
                        kcu.table_name AS foreign_table,
                        rel_kcu.table_name AS primary_table,
                        kcu.column_name AS foreign_column,
                        rel_kcu.column_name AS primary_column
                    FROM 
                        information_schema.table_constraints tco
                        JOIN information_schema.key_column_usage kcu
                        ON tco.constraint_schema = kcu.constraint_schema
                        AND tco.constraint_name = kcu.constraint_name
                        JOIN information_schema.referential_constraints rco
                        ON tco.constraint_schema = rco.constraint_schema
                        AND tco.constraint_name = rco.constraint_name
                        JOIN information_schema.key_column_usage rel_kcu
                        ON rco.unique_constraint_schema = rel_kcu.constraint_schema
                        AND rco.unique_constraint_name = rel_kcu.constraint_name
                        AND kcu.ordinal_position = rel_kcu.ordinal_position
                    WHERE tco.constraint_type = 'FOREIGN KEY' AND kcu.table_schema = 'public'
                    ORDER BY kcu.table_name, kcu.column_name;
                """))
                foreign_keys = foreign_keys_result.fetchall()

                prompt += f"\n\nHere is the information about the database tables:\n\n"
                for table in table_columns:
                    prompt += f"Table: {table[0]}, Column: {table[1]}, Data Type: {table[2]}\n"

                prompt += f"\nPrimary Keys:\n"
                for pk in primary_keys:
                    prompt += f"Table: {pk[0]}, Column: {pk[2]}, Constraint: {pk[1]}\n"

                prompt += f"\nForeign Keys:\n"
                for fk in foreign_keys:
                    prompt += f"Foreign Table: {fk[0]}, Foreign Column: {fk[2]}, Primary Table: {fk[1]}, Primary Column: {fk[3]}\n"

            prompt += f"""
                The query will be run against a Postgres database.

                CONTEXTUAL INSTRUCTIONS:
                - Understand the context of the request to ensure the SQL query correctly identifies and filters for the relevant entities.
                - Maintain context from previous questions and ensure the current query builds on previous results when needed.
                - Be mindful of pronouns and ambiguous terms, ensuring they are mapped to the correct entities and columns.

                SQL INSTRUCTIONS:
                - The query must be written specifically for PostgreSQL.
                - Do not use parameterized queries.
                - Only use the columns from the provided table information.
                - Try to make the query as efficient as possible.
                - However, it's important not to LIMIT the query when looking for data on more than one entity.
                -- DO NOT USE: SELECT track, release_date FROM most_streamed_spotify_songs_2024 WHERE LOWER(track) LIKE '%gypsy%' OR LOWER(track) LIKE '%a bar song (tipsy)%' AND release_date IS NOT NULL ORDER BY release_date ASC LIMIT 1;
                -- DO USE: SELECT track, release_date FROM most_streamed_spotify_songs_2024 WHERE LOWER(track) LIKE '%gypsy%' OR LOWER(track) LIKE '%a bar song (tipsy)%' AND release_date IS NOT NULL ORDER BY release_date ASC;
                - If the query looks like it is evaluating the possessive of a value, search for all variations.
                -- DO NOT USE: WHERE column_name = 'shaboozeys';
                -- DO USE: WHERE column_name = 'shaboozey' OR column_name = "shaboozey's" OR column_name = 'shaboozeys';
                - Round computation results to two decimal places.
                - Avoid symbols such as '?', '(', ')', '%', '$', ':' in comparisons.
                - Use 'LIKE' for string comparisons instead of '='.
                - Always use wildcard characters with 'LIKE' for partial matches (e.g., '%value%').
                -- DO NOT USE: WHERE column_name LIKE 'value';
                -- DO USE: WHERE column_name LIKE '%value%';
                - Write a SQL query adhering to general SQL standards.
                - Ensure the query does not exceed the LLM's context length.
                - Ensure the query is syntactically correct for PostgreSQL.
                - Format date comparisons as 'YYYY-MM-DD HH:MM:SS' and consider PostgreSQL's date functions.
                - Use case-insensitive comparisons (e.g., 'WHERE LOWER(column_name) LIKE ...').
                - Optimize the query to return only the necessary data using clauses like 'DISTINCT', 'WHERE', 'GROUP BY', 'ORDER BY', 'JOIN', 'UNION'.
                - Eliminate null values in all columns to ensure accurate results. Use 'WHERE column_name IS NOT NULL' for all relevant columns.
                - Do not use aggregate functions in the WHERE clause. Use HAVING for conditions involving aggregate functions or FILTER.
                - Ensure subqueries return only one row per group to avoid cardinality violation errors. Use 'DISTINCT ON' if necessary.
                - Explicitly cast fields to the appropriate data types when performing operations like ROUND, AVG, SUM, CORR, etc., to avoid type mismatch errors.
                
                RESPONSE INSTRUCTIONS:
                - Return ONLY the following key-value pairs: 'sql': "your sql query here" and 'in_domain': "boolean True or False if you think the query relates data in the database and will be answered with the SQL".
                - Only include the dictionary in the response. Don't include '```python' or '```' in the response.
                - The response should be a python dictionary that returns 'dict' when evaluated: type(eval(response)) == dict
            """

            return prompt
        except Exception as e:
            # logging.error(f"Error building SQL prompt: {e}")
            logging.error(f"Error building SQL prompt...")
            raise

    @disable_logging
    def _sql_openai_call(self, prompt: str) -> dict:
        """
        Make a call to the OpenAI API to generate SQL based on the prompt.

        Args:
            prompt (str): The constructed prompt string.

        Returns:
            dict: The generated SQL query and in_domain status.
        """
        try:
            response = client.chat.completions.create(
                model="gpt-4o",
                messages=[
                    {"role": "system", "content": "You are a helpful research assistant and a SQL expert for Postgres databases."},
                    {"role": "user", "content": prompt},
                ],
                temperature=0,
                max_tokens=1024,
                top_p=1,
                frequency_penalty=0,
                presence_penalty=0,
            )
            result = response.choices[0].message.content
            result = re.sub(r'```.*?```', '', result, flags=re.DOTALL)
            if isinstance(eval(result), dict) and 'sql' in eval(result) and 'in_domain' in eval(result):
                return eval(result)
            return None
        except Exception as e:
            # logging.error(f"Error in _sql_openai_call: {e}")
            logging.error(f"Error in _sql_openai_call...")
            return None

    def _build_generative_prompt(self, query: str, df: pd.DataFrame) -> str:
        """
        Build a prompt for generating a response based on the query and DataFrame.

        Args:
            query (str): The input query string.
            df (pd.DataFrame): The DataFrame containing the query results.

        Returns:
            str: The constructed prompt string for generative response.
        """
        try:
            prompt = f"""
            Based on the question:\n\n**{query}**\n\n the following data was found:\n\n
            SQL Data in JSON format: {df.to_json()}\n

            Do not make reference to the data sources themselves, only reference the data. For example, don't mention 'the data', 'JSON', 'SQL data', 'databases', etc.\n

            Unless specified in the query, don't show your work for mathematical calculations. Only provide the final answer.\n

            **This is a retrieval-augmented generation task, so it is critical that you only generate the answer based on the data provided in this prompt.**
            **If you need to make any assumptions, please state them clearly.**
            **If you think the data provided is insufficient to generate the answer, please state that as well.**
            """
            return prompt
        except Exception as e:
            # logging.error(f"Error building generative prompt: {e}")
            logging.error(f"Error building generative prompt...")
            raise

    @disable_logging
    def _generative_openai_call(self, prompt: str) -> str:
        """
        Make a call to the OpenAI API to generate a response based on the prompt.

        Args:
            prompt (str): The constructed prompt string.

        Returns:
            str: The generated response.
        """
        try:
            response = client.chat.completions.create(
                model="gpt-4o",
                messages=[
                    {"role": "system", "content": "You are a helpful research assistant and a SQL expert for Postgres databases."},
                    {"role": "user", "content": prompt},
                ],
                temperature=0,
                max_tokens=1024,
                top_p=1,
                frequency_penalty=0,
                presence_penalty=0,
            )
            return response.choices[0].message.content
        except Exception as e:
            # logging.error(f"Error generating response: {e}")
            logging.error(f"Error generating response...")
            return None


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.0.32'
DESCRIPTION = ''
LONG_DESCRIPTION = ''

setup(
    name="tabledai",
    version=VERSION,
    author="Vince Berry",
    author_email="vincent.berry11@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(exclude=['tabledai/create.py']),
    install_requires=['pandas', 'sqlalchemy', 'psycopg2-binary', 'python-dotenv', 'openai', 'chardet'],
    keywords=['python', 'ai', 'ml', 'nlp', 'sql', 'postgres'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)